/************************************************************************************
MODULE     :   LocBusSlave        AUTHOR    :   Tullio Grassi

DESCRIPTION :
The Local Bus interfaces different components on a board. This bus includes the data
lines, the address lines, WRITE_b (active low) and several CSn_b (active low
ChipSelect: CS1_b, CS2_b, for all the  slave devices).
The protocol foresees that when the master drives CS_b low for a particular slave,
the slave must decode the address on the Loc_Address lines, and read or write the 
Loc_Data bus depending of the WRITE_b direction.
On a write cycle, the master set WRITE_b = 0, set the address and data on the Local Bus,
and the slave will latch the data when CS_b = 0. 
The master keeps the data valid for the whole duration of the CS_b pulse. 
On a read cycle, the master set WRITE_b = 1, set the address on the Local Bus and
latch the data on the rising edge of CS_b, as the slave must prepare the data 
during the CS_b pulse. The CS_b pulse width depends on the slower slave.
A value of 300 ns is usually fine. When a slave is not selected, its data port is on Hi-Z.
The design is fully synchronous with clk. 
LocalBus lines are syncronized (this adds a few ticks of delay for edge detection).

!!!! When adding registers, remember that some of them are read-only,
some are write-only. Do not assign addresses already used !!!!!
**********************************************************************************/

`include "define.v"
module LocBusSlave  (      
// these ports connect to the LocalBus
  input  LocCS_b,  LocWRITE_b,       
  input  [`LB_Addr_Size-1:0]  LocAddr,  // The Address bus.
  inout  [`REG_WIDTH-1:0]     LocData,  // The bidir Data bus 
// these ports connect to the rest of the FPGA
  input clk,  rst, DLL_lock,       run_status,  BunchEr, 
        KChA, IdleA, ErrA,         KChB, IdleB, ErrB, 
        KChC, IdleC, ErrC,         KChD, IdleD, ErrD, 
  input [2:0] UnlockCnt,
  input [15:0] AccCntErr, AccR1sErr,
               RxClkCnt, RxBC0Cnt, TClkCnt, TxBC0Cnt,
  input [8:1]  InstantCntErr, InstantR1sErr,
               TPSpyFifoEmpty,  TPSpyFifoFull, 
  input [8:0]  TPSpyFIFO1_out,  TPSpyFIFO2_out,
               TPSpyFIFO3_out,  TPSpyFIFO4_out,
               TPSpyFIFO5_out,  TPSpyFIFO6_out,
               TPSpyFIFO7_out,  TPSpyFIFO8_out,
  input [4:1]  VitSpyFifoEmpty, VitSpyFifoFull,
  input [7:0]  LastTTCcommand,
  input [10:0] VitSpyFIFO1_out, VitSpyFIFO2_out,
               VitSpyFIFO3_out, VitSpyFIFO4_out,
  input [15:0] TTC_test0, TTC_test1, TTC_test2, TTC_test3,
               TTC_test4, TTC_test5, TTC_test6, TTC_test7,
               TTC_test8, TTC_test9, TTC_testA, TTC_testB,
               TTC_testC, TTC_testD, TTC_testE, TTC_testF,

  output  DisSODV, CounterModeSel, GateTTCcommands,  
          VME_Hard_rst, VME_Soft_rst, VME_VitResetN,
          VME_Start, VME_Stop, VME_LVDSout,
  output reg [`BCN_WIDTH-1:0] MaxBCN ='hDEB,  
  output reg [8:1] TPSpyFifoRd,
  output reg [4:1] VitSpyFifoRd,     
   // outputs for debug 
  output reg Tri,  
  output reg [2:0]  LB_STATE,       
  output reg [`REG_WIDTH-1:0] Dout,    
  output     [`REG_WIDTH-1:0] Din            );

  parameter 
  //  Parameters used for the STATEs of the LocBus FSM
  IDLE           =  0, // Waiting for the negative transition on LocCS_b
  RD_OR_WR       =  1, // check LocWRITE_b.  
  MASTER_WRITE   =  2, // If WR mode, then Store data
  MASTER_READ    =  3, // If RD mode, prepare data to be read
  END_OF_READ    =  4, // Latch data, release control lines, wait posedge of LocCS_b
  END_OF_WRITE   =  5, // Stop WR pulses and wait for posedge of LocCS_b
  END_CYCLE      =  6; // Set Data lines to HiZ (has to be after posedge)

//     INTERNAL SIGNALS
  reg  [`REG_WIDTH-1:0] CntrReg, StatusReg, Vitesse_StatReg, WriteOnlyCntrReg;
  reg  [`REG_WIDTH-1:0] SlbsID ='hB180; 
  reg  LocCS_Del_0, LocCS_Del_1, LocCS_Del_2; 
  
//  ACTUAL CODE STARTS HERE
  assign  DisSODV            =  CntrReg[1];
  assign  CounterModeSel     =  CntrReg[10];
  assign  GateTTCcommands    =  CntrReg[11];

  //  WriteOnlyCntrReg self-reset itself and creates positive pulses
  assign  VME_Hard_rst        =   WriteOnlyCntrReg[0];
  assign  VME_Start           =   WriteOnlyCntrReg[1];
  assign  VME_Stop            =   WriteOnlyCntrReg[2];
  assign  VME_Soft_rst        =   WriteOnlyCntrReg[3];
  assign  VME_VitResetN       =   WriteOnlyCntrReg[4];
  assign  VME_LVDSout         =   WriteOnlyCntrReg[5]; 

// forcing IOBUFfer to convice XST.  LocData goes stright to the pads while the rest of
//  the logic makes use of Din and Dout. NB: T=1 means Tristate.
//    _S_2  means Slow output, 2 mA drive (minimum, in case of bus collisions)
 IOBUF_S_2 LocDataBUF0  (.IO(LocData[0]),  .I(Dout[0]),  .O(Din[0]),  .T(Tri));
 IOBUF_S_2 LocDataBUF1  (.IO(LocData[1]),  .I(Dout[1]),  .O(Din[1]),  .T(Tri));
 IOBUF_S_2 LocDataBUF2  (.IO(LocData[2]),  .I(Dout[2]),  .O(Din[2]),  .T(Tri));
 IOBUF_S_2 LocDataBUF3  (.IO(LocData[3]),  .I(Dout[3]),  .O(Din[3]),  .T(Tri));
 IOBUF_S_2 LocDataBUF4  (.IO(LocData[4]),  .I(Dout[4]),  .O(Din[4]),  .T(Tri));
 IOBUF_S_2 LocDataBUF5  (.IO(LocData[5]),  .I(Dout[5]),  .O(Din[5]),  .T(Tri));
 IOBUF_S_2 LocDataBUF6  (.IO(LocData[6]),  .I(Dout[6]),  .O(Din[6]),  .T(Tri));
 IOBUF_S_2 LocDataBUF7  (.IO(LocData[7]),  .I(Dout[7]),  .O(Din[7]),  .T(Tri));
 IOBUF_S_2 LocDataBUF8  (.IO(LocData[8]),  .I(Dout[8]),  .O(Din[8]),  .T(Tri));
 IOBUF_S_2 LocDataBUF9  (.IO(LocData[9]),  .I(Dout[9]),  .O(Din[9]),  .T(Tri));
 IOBUF_S_2 LocDataBUF10 (.IO(LocData[10]), .I(Dout[10]), .O(Din[10]), .T(Tri));
 IOBUF_S_2 LocDataBUF11 (.IO(LocData[11]), .I(Dout[11]), .O(Din[11]), .T(Tri));
 IOBUF_S_2 LocDataBUF12 (.IO(LocData[12]), .I(Dout[12]), .O(Din[12]), .T(Tri));
 IOBUF_S_2 LocDataBUF13 (.IO(LocData[13]), .I(Dout[13]), .O(Din[13]), .T(Tri));
 IOBUF_S_2 LocDataBUF14 (.IO(LocData[14]), .I(Dout[14]), .O(Din[14]), .T(Tri));
 IOBUF_S_2 LocDataBUF15 (.IO(LocData[15]), .I(Dout[15]), .O(Din[15]), .T(Tri));

 always@(posedge clk)  begin  
   if(rst==1)   begin       //   RESET
         SlbsID <=  'hB180;     
   end  //   END of RESET 
   else  begin   // posedge clk       
      LocCS_Del_0  <=  LocCS_b;    // FF to avoid metastability
      LocCS_Del_1  <=  LocCS_Del_0;
      LocCS_Del_2  <=  LocCS_Del_1;  // total of 3Tck = 75 ns delay on CS
      StatusReg[0]                 <=  DLL_lock;
      StatusReg[1]                 <=  run_status;  
      StatusReg[2]                 <=  BunchEr;
      StatusReg[7:3]               <=  0;
      StatusReg[10:8]              <=  UnlockCnt;
	   StatusReg[`REG_WIDTH-1:11]   <=  0;  // unused portion of  reg
      Vitesse_StatReg              <=  {1'b0, KChD, IdleD, ErrD,
                                        1'b0, KChC, IdleC, ErrC,
                                        1'b0, KChB, IdleB, ErrB,
                                        1'b0, KChA, IdleA, ErrA};      

   //   some VITESSE RECEIVER STATUS SIGNALS
   // ERRn 	KCHn  IDLEn    Priority   Link Status
   // ------------------------------------------
   //   0     0     0          7	      Valid Data
   //   1     1     0          2	      Loss of Sync
   //   1     1     1          2	      Resync (includes IDLE)
    
    case(LB_STATE)
        
    IDLE:  // check for a negative edge on LocCS_b (digitally filtered)
        begin
          Dout      <=  'h BAD;  //mark a bad LocBus read timing
          Tri       <=  1;    // Data lines normally Hi-Z
          if(LocCS_Del_2 == 1 && LocCS_Del_1==0)  LB_STATE  <=  RD_OR_WR;
        end
            
    RD_OR_WR: //  decide whether it is a READ or a WRITE (master's point of view)
        begin
          if (LocWRITE_b==0)   LB_STATE <= MASTER_WRITE; 
          else  begin 
             LB_STATE <= MASTER_READ;  //  Read operation (master's point of view)
             Tri      <= 0;            //  Set Data lines toward LocalBus
            end  //  else
        end          
      //  ***********************************************   
      //     ********   W R I T E  operation  ********    
    MASTER_WRITE:  begin
       casez(LocAddr)    // decode which register is addressed                           
        'h0:      CntrReg           <=  Din[`REG_WIDTH-1:0];
        'h4:      SlbsID            <=  Din[`REG_WIDTH-1:0];     
        'h17:     MaxBCN            <=  Din[`REG_WIDTH-1:0];                            
        'h61:     WriteOnlyCntrReg  <=  Din[`REG_WIDTH-1:0];
        default:       ;
               //no action  if there is no such register or is read-only 
       endcase    //    (LocAddr)
       LB_STATE  <=  END_OF_WRITE;          
    end        //  end of MASTER_WRITE case

       //  ****************************************
       //    ******* R E A D  operation ********
    MASTER_READ:      begin
       casez(LocAddr)  // can use all Addr: 0 to 7F                                  
          // FIFO section: generate Fifo_RD pulses,
         // the data will be available on the next cycle
        'h31:   VitSpyFifoRd[1]    <=  1'b1;      
        'h32:   VitSpyFifoRd[2]    <=  1'b1;      
        'h33:   VitSpyFifoRd[3]    <=  1'b1;      
        'h34:   VitSpyFifoRd[4]    <=  1'b1;      
        'h35:   TPSpyFifoRd[1]     <=  1'b1;      
        'h36:   TPSpyFifoRd[2]     <=  1'b1;      
        'h37:   TPSpyFifoRd[3]     <=  1'b1;      
        'h38:   TPSpyFifoRd[4]     <=  1'b1;      
        'h39:   TPSpyFifoRd[5]     <=  1'b1;      
        'h3A:   TPSpyFifoRd[6]     <=  1'b1;      
        'h3B:   TPSpyFifoRd[7]     <=  1'b1;      
        'h3C:   TPSpyFifoRd[8]     <=  1'b1;      
        default:       ;
       endcase   //  (LocAddr)
       LB_STATE  <=  END_OF_READ;          
      end      //  MASTER_READ  case
            
    END_OF_READ:  begin                   
       // release all fifo read enables
       VitSpyFifoRd     <=  0;
       TPSpyFifoRd      <=  0;

       casez(LocAddr)  // here we actually latch the data outputs 
        'h0:  Dout <= CntrReg;
        'h1:  Dout <= 'hFFFFF;  //  this reg is for testing
        'h2:  Dout <= 'h0;      //  this reg is for testing
        'h3:  Dout <= `FPGAversionN;
        'h4:  Dout <= SlbsID;
        'hA:  Dout <= {InstantCntErr[8:1], InstantR1sErr[8:1]};
        'hB:  Dout <= AccCntErr;
        'hC:  Dout <= AccR1sErr;
        'h10: Dout <= StatusReg;
        'h17: Dout <= {4'b0, MaxBCN};
        'h18: Dout <= {8'b0, LastTTCcommand[7:0]}; // EvN_res not there
        'h20: Dout <= TTC_test0;   
        'h21: Dout <= TTC_test1;
        'h22: Dout <= TTC_test2;
        'h23: Dout <= TTC_test3;
        'h24: Dout <= TTC_test4;
        'h25: Dout <= TTC_test5;
        'h26: Dout <= TTC_test6;
        'h27: Dout <= TTC_test7;
        'h28: Dout <= TTC_test8;
        'h29: Dout <= TTC_test9;
        'h2A: Dout <= TTC_testA;
        'h2B: Dout <= TTC_testB;
        'h2C: Dout <= TTC_testC;
        'h2D: Dout <= TTC_testD;
        'h2E: Dout <= TTC_testE;
        'h2F: Dout <= TTC_testF;                           

        'h31: Dout <=  {4'b0, VitSpyFifoEmpty[1], VitSpyFIFO1_out};      
        'h32: Dout <=  {4'b0, VitSpyFifoEmpty[2], VitSpyFIFO2_out};      
        'h33: Dout <=  {4'b0, VitSpyFifoEmpty[3], VitSpyFIFO3_out};      
        'h34: Dout <=  {4'b0, VitSpyFifoEmpty[4], VitSpyFIFO4_out};  
            
        'h35: Dout <=  {6'b0, TPSpyFifoEmpty[1],  TPSpyFIFO1_out};      
        'h36: Dout <=  {6'b0, TPSpyFifoEmpty[2],  TPSpyFIFO2_out};     
        'h37: Dout <=  {6'b0, TPSpyFifoEmpty[3],  TPSpyFIFO3_out};    
        'h38: Dout <=  {6'b0, TPSpyFifoEmpty[4],  TPSpyFIFO4_out};     
        'h39: Dout <=  {6'b0, TPSpyFifoEmpty[5],  TPSpyFIFO5_out}; 
        'h3A: Dout <=  {6'b0, TPSpyFifoEmpty[6],  TPSpyFIFO6_out};   
        'h3B: Dout <=  {6'b0, TPSpyFifoEmpty[7],  TPSpyFIFO7_out}; 
        'h3C: Dout <=  {6'b0, TPSpyFifoEmpty[8],  TPSpyFIFO8_out};     

        'h3D: Dout <=  TClkCnt;   
        'h3E: Dout <=  RxClkCnt; 
        'h3F: Dout <=  TxBC0Cnt;     
        'h40: Dout <=  RxBC0Cnt; 
     
        'h60: Dout <= Vitesse_StatReg;
        'h70: Dout <= {7'b0, run_status, VitSpyFifoFull[4:1], VitSpyFifoEmpty[4:1]};
        'h71: Dout <= {TPSpyFifoFull[8:1], TPSpyFifoEmpty[8:1]};

        default:  Dout <=  'H DEAD; // if there is no such read-reg, answer: DEAD 
       endcase              
       if(LocCS_Del_1==1)    LB_STATE    <=  END_CYCLE;
    end   // END_OF_READ 
          
    END_OF_WRITE:
       begin    
          WriteOnlyCntrReg <=  0;   // self-reset to generate 1-tick pulses
          if(LocCS_Del_1==1)   LB_STATE    <=  END_CYCLE;
       end
          
    END_CYCLE:          
             LB_STATE      <= IDLE;
    
    default: LB_STATE    <=  IDLE;
    endcase    //  end the case related to the LB_STATE variable
   end      // else posedge clk
   end        // always
endmodule